/**
 # minimax.js
 # Implements Minimax Alpha-Beta-Pruning Algorithm to play connect4 game  
 #
 # @author Joachim Schmidt - joachim.schmidt@jschmidt-systemberatung.de
 # @copyright Copyright (C) 2022 Joachim Schmidt. All rights reserved.
 # @license	 http://www.gnu.org/licenses/gpl-2.0.html GNU/GPL
 #
 # change activity:
 */

let checkboard = [
  ['', '', '', '', '', '', ''],
  ['', '', '', '', '', '', ''],
  ['', '', '', '', '', '', ''],
  ['', '', '', '', '', '', ''],
  ['', '', '', '', '', '', ''],
  ['', '', '', '', '', '', '']
];

let maxDepth = 5;  

let scores = {
    ai: 1000000,
    human: -1000000,
    tie: 0
};
let human = 'r';
let ai = 'b';
let currentPlayer = human;

function getAvailableRow(col) {
    let row;
    for (let i = 0; i < 6; i++) {
        if (checkboard[i][col] == '') row = i;
    }
    return row;
}

function setOf(arr) {
    let newArr = [];
    for (let i of arr) {
        if (!newArr.includes(i)) newArr.push(i);
    }
    return newArr;
}

function same4(arr) {
    for (let i of arr) {
        if (i == '') return false;
    }
    if (arr.length != 4) console.log(`same4: arr length is not 4 - [${arr}]`);
    return (setOf(arr).length == 1);
}

function checkWin() {
    let winner;
    // Horizontal
    for (let row of checkboard) {
        for (let j = 0; j < 4; j++) {
            if (same4(row.slice(j, j + 4))) winner = row[j];
        }
    }
    // Vertical
    for (let i = 0; i < 7; i++) {
        let row = [];
        for (let r of checkboard) {
            row.push(r[i]);
        }
        for (let j = 0; j < 3; j++) {
            if (same4(row.slice(j, j + 4))) winner = row[j];
        }
    }
    // Diagonal
    for (let j = 0; j < 4; j++) {
        // upper diagonal
        for (let i = 3; i < 6; i++) {
            let d = [];
            for (let k = 0; k < 4; k++) {
                d.push(checkboard[i - k][j + k]);
            }
            if (same4(d)) {
                winner = d[0];
            }
        }
        // lower diagonal
        for (let i = 0; i < 3; i++) {
            let d = [];
            for (let k = 0; k < 4; k++) {
                d.push(checkboard[i + k][j + k]);
            }
            if (same4(d)) {
                winner = d[0];
            }
        }
    }
    let os = 0;
    for (let i = 0; i < 6; i++) {
        for (let j = 0; j < 7; j++) {
            if (checkboard[i][j] == '') os++;
        }
    }
    if (winner == null && os == 0) winner = 'tie';
    //console.log('winner = ' + winner);
    return winner;
}

function getAvailableRow(col) {
    let row;
    for (let i = 0; i < 6; i++) {
        if (checkboard[i][col] == '') row = i;
    }
    return row;
}

function evalArr(arr) {
    let aiP = 0;
    let humanP = 0;
    let emptyP = 0;
    for (let i of arr) {
        if (i == ai) aiP++;
        if (i == human) humanP++;
        if (i == "") emptyP++;
    }
    if (aiP == 3 && emptyP == 1) return 100;
    if (aiP == 2 && emptyP == 2) return 10;
    if (aiP == 1 && emptyP == 3) return 1;
    if (humanP == 3 && emptyP == 1) return -100;
    if (humanP == 2 && emptyP == 2) return -10;
    if (humanP == 1 && emptyP == 3) return -1;
    return 0;
}

function evalStaticBoard() {
	
	let score = 0;    
    // Horizontal
    for (let row of checkboard) {
        for (let j = 0; j < 4; j++) {
            score += evalArr(row.slice(j, j + 4));
        }
    }
    // Vertical
    for (let i = 0; i < 7; i++) {
        let row = [];
        for (let r of checkboard) {
            row.push(r[i]);
        }
        for (let j = 0; j < 3; j++) {
            score += evalArr(row.slice(j, j + 4));
        }
    }
    // Diagonal
    for (let j = 0; j < 4; j++) {
        // upper diagonal
        for (let i = 3; i < 6; i++) {
            let d = [];
            for (let k = 0; k < 4; k++) {
                d.push(checkboard[i - k][j + k]);
            }
            score += evalArr(d);
        }
        // lower diagonal
        for (let i = 0; i < 3; i++) {
            let d = [];
            for (let k = 0; k < 4; k++) {
                d.push(checkboard[i + k][j + k]);
            }
            score += evalArr(d);
        }
    }
    return score;
}

function aiMove() {
    let bestScore = -Infinity;
    let bestMove;
    for (let i = 0; i < 7; i++) {
        let j = getAvailableRow(i);
        if (j != null && checkboard[j][i] == "") {
            checkboard[j][i] = ai;
            let score;
            if (false) {
                score = minimax(checkboard, maxDepth, false);
            } else {
                score = minimaxABP(checkboard, maxDepth, false, -Infinity, Infinity);
            }
            checkboard[j][i] = "";
            if (score >= bestScore) {
                bestScore = score;
                bestMove = {
                    i,
                    j
                };
            }
        }
    }
    if (bestMove != null) {
        checkboard[bestMove.j][bestMove.i] = ai;
        return bestMove.i;
    }
}

function minimax(checkboard, depth, isMaximizing) {
    
    let nMoves = maxDepth - depth + 1;
    let r = checkWin();
    if (r != null) {
        let s;
        if (r == human) s = scores.human;
        else if (r == ai) s = scores.ai;
        else if (r == "tie") s = scores.tie;
        return s - 20 * nMoves;
    }
    if (depth == 0) {
        return evalStaticBoard() - 20 * nMoves;
    }
    if (isMaximizing) {
        let bestScore = -Infinity;
        for (let i = 0; i < 7; i++) {
            let j = getAvailableRow(i);
            if (j != null && checkboard[j][i] == "") {
                checkboard[j][i] = ai;
                let score = minimax(checkboard, depth - 1, false);
                checkboard[j][i] = "";
                bestScore = Math.max(bestScore, score);
            }
        }
        return bestScore;
    } else {
        let bestScore = Infinity;
        for (let i = 0; i < 7; i++) {
            let j = getAvailableRow(i);
            if (j != null && checkboard[j][i] == "") {
                checkboard[j][i] = human;
                let score = minimax(checkboard, depth - 1, true);
                checkboard[j][i] = "";
                bestScore = Math.min(bestScore, score);
            }
        }
        return bestScore;
    }
}

function minimaxABP(checkboard, depth, isMaximizing, alpha, beta) {
    let nMoves = maxDepth - depth + 1;
    let r = checkWin();
    if (r != null) {
        let s;
        if (r == human) s = scores.human;
        else if (r == ai) s = scores.ai;
        else if (r == "tie") s = scores.tie;
        return s - 20 * nMoves;
    }
    if (depth == 0) {
        return evalStaticBoard() - 20 * nMoves;
    }
    if (isMaximizing) {
        let bestScore = -Infinity;
        for (let i = 0; i < 7; i++) {
            let j = getAvailableRow(i);
            if (j != null && checkboard[j][i] == "") {
                checkboard[j][i] = ai;
                let score = minimaxABP(checkboard, depth - 1, false, alpha, beta);
                checkboard[j][i] = "";
                bestScore = Math.max(bestScore, score);
                alpha = Math.max(alpha, bestScore);
                if (beta <= alpha) break;
            }
        }
        return bestScore;
    } else {
        let bestScore = Infinity;
        for (let i = 0; i < 7; i++) {
            let j = getAvailableRow(i);
            if (j != null && checkboard[j][i] == "") {
                checkboard[j][i] = human;
                let score = minimaxABP(checkboard, depth - 1, true, alpha, beta);
                checkboard[j][i] = "";
                bestScore = Math.min(bestScore, score);
                beta = Math.min(beta, bestScore);
                if (beta <= alpha) break;
            }
        }
        return bestScore;
    }
}